# pseudom209 : The base Class of M209 : we use only Pins

import string
import sys
import random
import math
from   bib_util import *

# ===================================
def IC( f ):
    N = 0 ; total = 0 ; ic = 0
    for i in range(26):
        #if ( f[i] > 1):
        total = total + f[i]*(f[i]-1)
        N += f[i]
    if N > 1:
        ic =  total / float(N * (N-1))
    return ic
# ===================================
class Wheel:
    def __init__(self,alphabet,shift):
        self.alphabet = alphabet
        self.lg = len(alphabet)
        self.pins = []
        for i in range(self.lg):
            self.pins.append( 0 )
        self.position = 0
        self.shiftWindow = shift

    def setPinsON(self,car):
        i = string.index(self.alphabet,car)
        self.pins[i] = 1

    def setPinsOFF(self,car):
        i = string.index(self.alphabet,car)
        self.pins[i] = 0
        
    def getAPins(self,car):
        i = string.index(self.alphabet,car)
        return self.pins[i]

    def getShift(self):
        return self.shiftWindow

    def turnClockwise(self):
        self.position = (self.position + 1 ) % self.lg

    def turnCounterClockwise(self):
        self.position = (self.position + self.lg - 1 ) % self.lg 

    def getLg( self):
        return len( self.alphabet )

    def getPosition(self):
        return self.position

    def setPosition(self, newPosition ):
        self.position = newPosition

    def getExternalKey(self):
        return self.alphabet[ self.position ]

    def getExternalShiftedKey(self):
        return self.alphabet[(self.position + self.shiftWindow)%self.lg]

    def getInternalKey(self):
        ch = ""
        for i in range(self.lg):
            if self.pins[i] :
                ch = ch + self.alphabet[i]
            else:
                ch = ch + "_"
        return ch

    def isActive(self):
        pos_active = ( self.position + self.shiftWindow ) % self.lg
        if self.pins[ pos_active ] != 0:
            return 1
        else:
            return 0

    def zero(self):
        for i in range(self.lg):
            self.pins[i] = 0

    def info(self):
        print "Lenght       : ", self.lg
        print "Alphabet     : ", self.alphabet
        print "Shitf        : ", self.shiftWindow
        print "Pins         : ", self.pins
        print "Internal Key : ", self.getInternalKey()
        print "External key : ", self.getExternalKey()
        print "Position: ", self.position, " Actif : ", self.isActive()

class Hagelin:
    def __init__( self, debug ):
        self.debug = debug
        self.wheels = []
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVWXYZ",15 ) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVXYZ",14) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVX",13) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTU",12) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRS",11) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQ",10  ) )
        self.nbOfWheels = len( self.wheels )
        self.counter  = 0

    def setDebug(self,debug):
        self.debug = debug

    def getDebug(self):
        return self.debug

    def setExternalKey( self, key ):
        for i in range( len(key) ):
            non_trouve = 1
            lettre_cle = key[i]
            lg = self.wheels[i].getLg()
            for j in range (lg):
                if lettre_cle == self.wheels[i].getExternalKey():
                    non_trouve = 0                  
                    break
                self.stepAWheel(i)
            if non_trouve: 
                print "**** Error: invalid external key  **** "; exit

    def setPinsON(self,laroue,char):
        self.wheels[laroue].setPinsON(char)

    def setPinsOFF(self,laroue,char):
        self.wheels[laroue].setPinsOFF(char)

    def setAllPins(self,laroue,ch):
        for i in range(len(ch)):
            if ch[i] == "1":
                self.setPinsON(laroue,
                self.wheels[laroue].alphabet[i] )
            else:
                self.setPinsOFF(laroue,
                self.wheels[laroue].alphabet[i] )

    def setAllPinsAllW(self,ch):
        self.setAllPins(0, ch[:26] )
        self.setAllPins(1, ch[26:51] )
        self.setAllPins(2, ch[51:74] )
        self.setAllPins(3, ch[74:95] )
        self.setAllPins(4, ch[95:114] )
        self.setAllPins(5, ch[114:] )

    def getPins(self,laroue,char):
        return self.wheels[laroue].getAPins(char)

    def stepAWheel(self, aWheel ):
        self.wheels[ aWheel ].turnClockwise()

    def step(self):
        for i in range( self.nbOfWheels ):      
            self.wheels[i].turnClockwise()
        self.counter = self.counter + 1

    def getPosition( self ):
        return self.counter

    def getPinsOfAWheel( self, aWheel ):
        return self.wheels[ aWheel ].getInternalKey()

    def stepBack(self):
        for i in range( self.nbOfWheels ):      
            self.wheels[i].turnCounterClockwise()
        self.counter = self.counter - 1
        if ( self.counter < 0 ):
            self.counter = 999

    def rewind(self):
        while ( self.counter != 0 ):
            self.stepBack()

    def getExternalKey( self ):
        ch = ""
        for i in range( self.nbOfWheels ):
            ch = ch + self.wheels[i].getExternalKey()
        return ch

    def i64(self):
        index = 0
        for j in range( self.nbOfWheels ):
            if self.wheels[ j ].isActive():
                index += [1,2,4,8,16,32][j]
        self.step()
        #if self.debug: print index,
        return index

    def getNbOfWheels(self):
        return self.nbOfWheels

    def toggleDebug( self ):
        if self.debug: self.debug = 0
        else: self.debug = 1

    def info(self):
        print "Lugs    : ", self.lugs
        print "Overlaps: ", self.over
        print "Actives Pins : ", 
        for i in range( self.nbOfWheels ):
            self.wheels[i].info()
#================================================
def cali64( e, lg ):
    lesIdx = []
    for i in range(lg):
        lesIdx.append( e.i64() )
    return lesIdx

def getAlpha( msg, lesIdx ):
    lesAlpha = []
    for i in range(64): lesAlpha.append( [0]*27 )
    for i in range(len(msg)):
        lettre = string.index(string.ascii_uppercase, msg[i])    
        print lettre,
        lesAlpha[ lesIdx[i] ][ lettre ] += 1
    return lesAlpha

def icBy64Box( e, msg ):
    lesAlpha = []
    lg = len(msg)
    for i in range(64): lesAlpha.append( [0]*27 )
    for i in range(lg):
        lettre = string.index(string.ascii_uppercase, msg[i])    
        idx = e.i64()
        lesAlpha[ idx ][ lettre ] += 1
        lesAlpha[ idx ][ 26 ] += 1
    ic_weighted = 0
    for i in range(64):
        ic = IC( lesAlpha[i] )
        weight = lesAlpha[i][26]
        if e.getDebug():
            if lesAlpha[i][26] != 0:
                print i, lesAlpha[i], ic, weight
        ic_weighted += ic * ( lesAlpha[i][26] / float(lg) )
    return ic_weighted, lesAlpha

# ==== on fait juste un teste de chiffrement
if __name__ == "__main__":
    e = Hagelin (0)
#          01234567890123456789012345
    Pins="11010001101011000011011000"+\
         "0001101001110010011010100"+\
         "11000011010111000111111"+\
         "001011011000110100111"+\
         "0101110110001101001"+\
         "11010001001001101"
    e.setAllPinsAllW(Pins)
    e.rewind()
    lesIdx =  cali64( e, 10 )
    print lesIdx
    #print chiffre(e,"AAAAAAAAAAAAAAAAAAAAAAAAAA")
    #print getAlpha ( "AAAAAAAAAA" , lesIdx)
    e.rewind()
    [ic, tb] = icBy64Box( e, "AQIOJSDIQOSJDJQSIODJOQSJDQSIJDOIQJSDIOJQSDJOQSJDQSOIDJOQISDQOISJDOIUIOAOJCOJIAZYHNBCBFAUYZEIIAZEPOOLKSIUDDJ" )
    print ">>>>>ic: ", ic
    #print tb
